/* ============================= C MeatAxe ==================================
   File:        $Id $
   Comment:     Input/output of integers.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysReadLong "Read long integers"
!synopsis 
    int SysReadLong(FILE *f, long *buf, int n);
!arg f
    File to read from.
!arg buf
    Pointer to buffer.
!arg n
    Number of integers to read.
!return
    Number of integers that were actually read, or $/1$ on error.
!description
    This function reads |n| long integers from the file |f| into the array 
    |buf|. |buf| must point to a memory area of at least |n*sizeof(long)| 
    bytes and |f| must be open for reading. the return value indicates how
    manu integers have actually been read. This number may be less than
    |n| because the end of file was encountered while reading. A negative
    return value indicates a file i/o error.
    
    This function expects that the numbers in the file are 4-byte integers 
    in `little-endian' format, i.e. the least significant byte first.
    Using a machine-independent data format makes {\MeatAxe} data files 
    more portable, but there are also some disadvantages:
    \begin{enumerate}
    \item 
	The conversion to and from machine-independent format involves several
	arithmetic operations for each number read/written.
    \item
	The highest number which can be read/written is $2^{32}-1$, even on
	64 bit machines.
    \end{enumerate}
!seealso SysWriteLong
-------------------------------------------------------------------------- }}}*/

int SysReadLong(FILE *f, long *buf, int n)

{
    unsigned char a[4];
    int nread;

    /* Check the arguments
       ------------------- */
    if (f == NULL || buf == NULL || n < 0)
    {
	MTX_ERROR3("Invalid arguments (f=%x,buf=%x,n=%d)",(int)f,(int)buf,n);
	return -1;
    }

    /* Read <n> long integers
       ---------------------- */
    if (Mtx_IsX86)
    {
	nread = (int)fread((char *)buf,sizeof(long),(size_t)n,f);
	/* printf("nread=%d, error=%d, eof=%d, GetLastError()=%d\n",
	   nread,ferror(f),feof(f),GetLastError());
	*/
    }
    else
    {
    	for (nread = 0; nread < n; ++nread)
    	{
	    if (fread(a,1,4,f) != 4) break;
	    *buf = ((unsigned long)a[0]|((unsigned long)a[1] << 8)|
		((unsigned long)a[2] << 16)|((long)(char)a[3] << 24));
	    ++buf;
    	}
    }

    /* Check for errors
       ---------------- */
    if (ferror(f) && !feof(f))
    {
	MTX_ERROR("Read failed: %S");
	return -1;
    }
    return nread;
}



/*{{{ --------------------------------------------------------------------------
!function SysWriteLong "Write long integers"
!synopsis 
    int SysWriteLong(FILE *f, const long *buf, int n);
!arg f
    File to write to.
!arg buf
    Pointer to buffer.
!arg n
    Number of integers to write.
!return
    Number of integers that were written, or $-1$ on error.
!description
    This function writes |n| long integers from the the array |buf| to the 
    file |f|. |buf| must point to a memory area of at least |n*sizeof(long)| 
    bytes and |f| must be open for writing. The numbers are written in a 
    machine-independent format which can be read by |SysReadLong()|.

!seealso SysReadLong
-------------------------------------------------------------------------- }}}*/

int SysWriteLong(FILE *f, const long *buf, int n)

{
    unsigned char a[4];
    int nwritten;

    if (Mtx_IsX86)
	nwritten = fwrite((char *)buf,sizeof(long),n,f);
    else
    {
    	for (nwritten = 0; nwritten < n; ++nwritten)
    	{
    	    a[0] = (unsigned char) *buf;
    	    a[1] = (unsigned char) (*buf >> 8);
    	    a[2] = (unsigned char) (*buf >> 16);
    	    a[3] = (unsigned char) (*buf >> 24);
    	    if (fwrite(a,1,4,f) != 4) break;
	    ++buf;
    	}
    }
    return nwritten;
}

